"evaluate"
Object subclass: #PPFailure
instanceVariableNames: 
	' message position '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPFailure comment: 'The failure object in PetitParser. It is the only class that responds to #isPetitFailure with true. It contains an error message and a position of the occurrence of the failure.

Instance Variables:
	message    <String>    The error message of this failure.
	position    <Integer>    The position of this failure in the input stream.
'!
"evaluate" 
Object subclass: #PPMemento
instanceVariableNames: 
	' result count position '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate" 
PPMemento comment: 'PPMemento is an internal class used by PPMemoizedParser to cache results and detect left-recursive calls.

Instance Variables:
	result    <Object>    The cached result.
	count    <Integer>    The number of recursive cycles followed.
	position    <Integer>    The position of the cached result in the input stream.
'!
"evaluate"
Object subclass: #PPToken
instanceVariableNames: 
	' collection start stop '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPToken comment: 'PPToken represents a parsed part of the input stream. Contrary to a simple String it remembers where it came from, the original collection and its start and stop position.

Instance Variables:
	collection    <SequenceableCollection>    The collection this token comes from.
	start    <Integer>    The start position in the collection.
	stop    <Integer>    The stop position in the collection.
'!
"evaluate"  
ReadStream subclass: #PPStream
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPStream comment: 'A positional stream implementation used for parsing. It overrides some methods for optimization reasons.
'!
!PPStream methodsFor: #converting!   
asPetitStream
	^ self
!   !
!PPStream methodsFor: #accessing!  
next: anInteger 
	"Answer up to anInteger elements of my collection. Overridden for efficiency."

	| answer endPosition |
	endPosition := position + anInteger min: readLimit.
	answer := collection copyFrom: position + 1 to: endPosition.
	position := endPosition.
	^ answer
!  !
!PPStream methodsFor: #accessing!  
peek
	"An improved version of peek, that is slightly faster than the built in version."

	^ self atEnd ifFalse: [ collection at: position + 1 ]
!   !
!PPStream methodsFor: #accessing!  
position: anInteger
	"The receiver does not check for invalid arguments passed to this method, as it is solely used with valid indexes for backtracking."

	position := anInteger
! !
!PPStream methodsFor: #printing!   
printOn: aStream
	aStream 
		nextPutAll: (collection copyFrom: 1 to: position);
		nextPutAll: '·';
		nextPutAll: (collection copyFrom: position + 1 to: readLimit)
!  !
!PPStream methods! 
uncheckedPeek
	^collection at: position + 1
! !
!PPMemento class methodsFor: #'instance creation'! 
new
	^ self basicNew initialize
! !
!PPMemento methodsFor: #'accessing-readonly'!  
count
	^ count
!  !
!PPMemento methodsFor: #actions!   
increment
	count := count + 1
!   !
!PPMemento methodsFor: #initialization!
initialize
	count := 0
	
!   !
!PPMemento methodsFor: #accessing! 
position
	^ position
!    !
!PPMemento methodsFor: #accessing! 
position: anInteger
	position := anInteger
!  !
!PPMemento methodsFor: #accessing! 
result
	^ result
!    !
!PPMemento methodsFor: #accessing! 
result: anObject
	result := anObject
!    !
!PPToken class methodsFor: #'instance creation'!   
new
	self error: 'Token can only be created using a dedicated constructor.'
! !
!PPToken class methodsFor: #'instance creation'!   
on: aSequenceableCollection
	^ self on: aSequenceableCollection start: 1 stop: aSequenceableCollection size
! !
!PPToken class methodsFor: #'instance creation'!   
on: aSequenceableCollection start: aStartInteger stop: aStopInteger
	^ self basicNew 
		initializeOn: aSequenceableCollection
		start: aStartInteger stop: aStopInteger
!   !
!PPToken methodsFor: #comparing!   
= anObject
	^ self class = anObject class and: [ self value = anObject value ]
!  !
!PPToken methodsFor: #comparing!   
hash
	^ self value hash
! !
!PPToken methodsFor: #accessing!   
collection
	"Answer the underlying collection of this token."

	^ collection
!  !
!PPToken methodsFor: #accessing!   
size
	"Answer the size of this token."

	^ stop - start + 1
!   !
!PPToken methodsFor: #accessing!   
start
	"Answer the start position of this token in the underlying collection."

	^ start
!  !
!PPToken methodsFor: #accessing!   
stop
	"Answer the stop position of this token in the underlying collection."
	
	^ stop
!    !
!PPToken methodsFor: #accessing!   
value
	"Answer the contents of this token."

	^ collection copyFrom: start to: stop
!   !
!PPToken methodsFor: #querying!
column
	"Answer the column number of this token in the underlying collection."
	
	| position |
	position := 0.
	(self newline , [ :stream |
		start <= stream position
			ifTrue: [ ^ start - position ].
		position := stream position ] asParser
		/ #any asParser) star
			parse: collection.
	^ start - position
!  !
!PPToken methodsFor: #querying!
line
	"Answer the line number of this token in the underlying collection."
	
	| line |
	line := 1.
	(self newline , [ :stream |
		start <= stream position
			ifTrue: [ ^ line ].
		line := line + 1 ] asParser
		/ #any asParser) star
			parse: collection.
	^ line
! !
!PPToken methodsFor: #copying! 
copyFrom: aStartInteger to: aStopInteger
	^ self class on: collection start: start + aStartInteger - 1 stop: stop + aStopInteger - 3
!    !
!PPToken methodsFor: #initialization!  
initializeOn: aSequenceableCollection start: aStartInteger stop: aStopInteger
	collection := aSequenceableCollection.
	start := aStartInteger.
	stop := aStopInteger
!  !
!PPToken methodsFor: #private! 
newline
	"Parser a platform independent newline sequence. LF: Unix, CR+LF: Windows, and CR: Apple."

	^ (Character lf asParser)
	/ (Character cr asParser , Character lf asParser optional)
!  !
!PPToken methodsFor: #printing!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(; nextPutAll: self value; nextPut: $)
!    !
!PPFailure class methodsFor: #'instance creation'! 
message: aString at: anInteger
	^ self basicNew initializeMessage: aString at: anInteger
!    !
!PPFailure methodsFor: #initialization!
initializeMessage: aString at: anInteger
	message := aString.
	position := anInteger
!   !
!PPFailure methodsFor: #testing!   
isPetitFailure
	"I am the only class that should implement this method to return true."

	^ true
!  !
!PPFailure methodsFor: #accessing! 
message
	"Answer a human readable error message of this parse failure."
	
	^ message
!  !
!PPFailure methodsFor: #accessing! 
position
	"Answer the position in the source string that caused this parse failure."

	^ position
! !
!PPFailure methodsFor: #printing!  
printOn: aStream
	aStream nextPutAll: self message; nextPutAll: ' at '; print: position
! !
