"evaluate"
PPDelegateParser subclass: #PPExpressionParser
instanceVariableNames: 
	' operators '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPExpressionParser comment: 'A PPExpressionParser is a parser to conveniently define an expression grammar with prefix, postfix, and left- and right-associative infix operators.

The following code initializes a parser for arithmetic expressions. First we instantiate an expression parser, a simple parser for expressions in parenthesis and a simple parser for integer numbers.

	expression := PPExpressionParser new.
	parens := $( asParser token trim , expression , $) asParser token trim 
		==> [ :nodes | nodes second ].
	integer := #digit asParser plus token trim
		==> [ :token | token value asInteger ].
	
Then we define on what term the expression grammar is built on:

	expression term: parens / integer.
	
Finally we define the operator-groups in descending precedence. Note, that the action blocks receive both, the terms and the parsed operator in the order they appear in the parsed input. 
	
	expression
		group: [ :g |
			g prefix: $- asParser token trim do: [ :op :a | a negated ] ];
		group: [ :g |
			g postfix: ''++'' asParser token trim do: [ :a :op | a + 1 ].
			g postfix: ''--'' asParser token trim do: [ :a :op | a - 1 ] ];
		group: [ :g |
			g right: $^ asParser token trim do: [ :a :op :b | a raisedTo: b ] ];
		group: [ :g |
			g left: $* asParser token trim do: [ :a :op :b | a * b ].
			g left: $/ asParser token trim do: [ :a :op :b | a / b ] ];
		group: [ :g |
			g left: $+ asParser token trim do: [ :a :op :b | a + b ].
			g left: $- asParser token trim do: [ :a :op :b | a - b ] ].
		
After evaluating the above code the ''expression'' is an efficient parser that evaluates examples like:

	expression parse: ''-8++''.
	expression parse: ''1+2*3''.
	expression parse: ''1*2+3''.
	expression parse: ''(1+2)*3''.
	expression parse: ''8/4/2''.
	expression parse: ''8/(4/2)''.
	expression parse: ''2^2^3''.
	expression parse: ''(2^2)^3''.
	
Instance Variables:
	operators	<Dictionary>	The operators defined in the current group.
'!
"evaluate" 
Object subclass: #PPCharSetPredicate
instanceVariableNames: 
	' block classification '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate" 
PPDelegateParser subclass: #PPCompositeParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate" 
PPCompositeParser comment: 'A PPCompositeParser is composed parser built from various primitive parsers. 

Every production in the receiver is specified as a method that returns its parser. Note that every production requires an instance variable of the same name, otherwise the production is not cached and cannot be used in recursive grammars. Productions should refer to each other by reading the respective inst-var. Note: these inst-vars are typically not written, as the assignment happens in the initialize method using reflection.

The start production is defined in the method start. It is aliased to the inst-var parser defined in the superclass of PPCompositeParser.
'!
"evaluate" 
PPParser subclass: #PPUnresolvedParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPUnresolvedParser comment: 'This is a temporary placeholder or forward reference to a parser that has not been defined yet. If everything goes well it will eventually be replaced with the real parser instance.
'!
!PPExpressionParser methodsFor: #private! 
build: aParser left: aChoiceParser
	^ (aParser separatedBy: aChoiceParser) foldLeft: [ :a :op :b | op first value: a value: op second value: b ]
! !
!PPExpressionParser methodsFor: #private!  
build: aParser postfix: aChoiceParser
	^ aParser , aChoiceParser star map: [ :term :ops | ops inject: term into: [ :result :operator | operator first value: result value: operator second ] ]
!   !
!PPExpressionParser methodsFor: #private!  
build: aParser prefix: aChoiceParser
	^ aChoiceParser star , aParser map: [ :ops :term | ops reversed inject: term into: [ :result :operator | operator first value: operator second value: result ] ]
!   !
!PPExpressionParser methodsFor: #private!  
build: aParser right: aChoiceParser
	^ (aParser separatedBy: aChoiceParser) foldRight: [ :a :op :b | op first value: a value: op second value: b ]
!   !
!PPExpressionParser methodsFor: #private!  
buildOn: aParser
	^self buildSelectors inject: aParser into: [:term :selector | | list |
		list := operators at: selector ifAbsent: [#()].
		list isEmpty ifTrue: [term] ifFalse: [
			self
				perform: selector
				with: term
				with: (list size = 1
					ifTrue: [list first first
						==> [:operator | Array with: list first second with: operator]]
					ifFalse: [
						list
							inject: PPChoiceParser new
							into: [:choice :each | choice
								/ (each first
									==> [:operator | Array with: each second with: operator])]])]]
!    !
!PPExpressionParser methodsFor: #private!  
buildSelectors
	^ #(build:prefix: build:postfix: build:right: build:left:)
!   !
!PPExpressionParser methodsFor: #private!  
operator: aSymbol parser: aParser do: aBlock
	parser isNil
		ifTrue: [ ^ self error: 'You did not specify a term when creating the receiver.' ].
	operators isNil
		ifTrue: [ ^ self error: 'Use #group: to define precedence groups in descending order.' ].
	(operators at: aSymbol ifAbsentPut: [ OrderedCollection new ])
		addLast: (Array with: aParser asParser with: aBlock)
! !
!PPExpressionParser methodsFor: #specifying!   
group: aOneArgumentBlock
	"Defines a priority group by evaluating aOneArgumentBlock."
	
	operators := Dictionary new.
	parser := [ 
		aOneArgumentBlock value: self.
	 	self buildOn: parser ]
			ensure: [ operators := nil ]
!   !
!PPExpressionParser methodsFor: #specifying!   
left: aParser do: aThreeArgumentBlock
	"Define an operator aParser that is left-associative. Evaluate aThreeArgumentBlock with the first term, the operator, and the second term."
	
	self operator: #build:left: parser: aParser do: aThreeArgumentBlock
!    !
!PPExpressionParser methodsFor: #specifying!   
postfix: aParser do: aTwoArgumentBlock
	"Define a postfix operator aParser. Evaluate aTwoArgumentBlock with the term and the operator."

	self operator: #build:postfix: parser: aParser do: aTwoArgumentBlock
!   !
!PPExpressionParser methodsFor: #specifying!   
prefix: aParser do: aTwoArgumentBlock
	"Define a prefix operator aParser. Evaluate aTwoArgumentBlock with the operator and the term."

	self operator: #build:prefix: parser: aParser do: aTwoArgumentBlock
!  !
!PPExpressionParser methodsFor: #specifying!   
right: aParser do: aThreeArgumentBlock
	"Define an operator aParser that is right-associative. Evaluate aThreeArgumentBlock with the first term, the operator, and the second term."
	
	self operator: #build:right: parser: aParser do: aThreeArgumentBlock
! !
!PPExpressionParser methodsFor: #specifying!   
term: aParser
	"Defines the initial term aParser of the receiver."
	
	parser isNil
		ifTrue: [ parser := aParser ]
		ifFalse: [ self error: 'Unable to redefine the term.' ]
! !
!PPCharSetPredicate class methodsFor: #'instance creation'!
on: aBlock
	^ self basicNew initializeOn: aBlock
! !
!PPCharSetPredicate methodsFor: #initialization!   
initializeOn: aBlock
	block := aBlock.
	classification := Array new: 255.
	1 to: classification size do: [ :index |
		classification at: index put: (block
			value: (Character value: index)) ]
! !
!PPCharSetPredicate methodsFor: #evaluating!   
value: aCharacter
	| index |
	index := aCharacter asInteger.
	index == 0
		ifTrue: [ ^ block value: aCharacter ].
	index > 255
		ifTrue: [ ^ block value: aCharacter ].
	^ classification at: index
!  !
!PPCompositeParser class methodsFor: #accessing!   
ignoredNames
	"Answer a collection of instance-variables that should not be automatically initialized with productions, but that are used internal to the composite parser."

	^ PPCompositeParser allInstVarNames
!   !
!PPCompositeParser class methodsFor: #accessing!   
startSymbol
	"Answer the method that represents the default start symbol."

	^ #start
!    !
!PPCompositeParser class methodsFor: #'instance creation'! 
new
	"Answer a new parser starting at the default start symbol."

	^ self newStartingAt: self startSymbol
!    !
!PPCompositeParser class methodsFor: #'instance creation'! 
newStartingAt: aSymbol
	"Answer a new parser starting at aSymbol."

	^ self basicNew initializeStartingAt: aSymbol
!   !
!PPCompositeParser class methodsFor: #parsing! 
parse: anObject
	^ self parse: anObject startingAt: self startSymbol
! !
!PPCompositeParser class methodsFor: #parsing! 
parse: anObject onError: aBlock
	^ self parse: anObject startingAt: self startSymbol onError: aBlock
! !
!PPCompositeParser class methodsFor: #parsing! 
parse: anObject startingAt: aSymbol
	^ (self newStartingAt: aSymbol) parse: anObject
! !
!PPCompositeParser class methodsFor: #parsing! 
parse: anObject startingAt: aSymbol onError: aBlock
	^ (self newStartingAt: aSymbol) parse: anObject onError: aBlock
! !
!PPCompositeParser methodsFor: #initialization!
initializeStartingAt: aSymbol
	| allVariableNames ignoredVariableNames productionIndexesAndNames |
	self initialize.	

	"find all the productions that need to be initialized"
	allVariableNames := self class allInstVarNames
		collect: [ :each | each asSymbol ].
	ignoredVariableNames := self class ignoredNames
		collect: [ :each | each asSymbol ].
	productionIndexesAndNames := ((1 to: self class instSize)
		collect: [ :index | index -> (allVariableNames at: index) ])
		reject: [ :assoc | ignoredVariableNames includes: assoc value ].
	
	"initialize productions with an undefined parser to be replaced later"
	parser := PPUnresolvedParser named: aSymbol.
	productionIndexesAndNames do: [ :assoc |
		self instVarAt: assoc key put: (PPUnresolvedParser named: assoc value) ].
	parser def: (self perform: aSymbol).
	
	"resolve unresolved parsers with their actual implementation"
	productionIndexesAndNames do: [ :assoc |
		(self respondsTo: assoc value)
			ifFalse: [ self error: 'Unable to initialize ' , assoc value printString ]
			ifTrue: [ (self instVarAt: assoc key) def: (self perform: assoc value) ] ]
! !
!PPCompositeParser methodsFor: #querying!  
productionAt: aSymbol
	"Answer the production named aSymbol."
	
	^ self productionAt: aSymbol ifAbsent: [ nil ]
!  !
!PPCompositeParser methodsFor: #querying!  
productionAt: aSymbol ifAbsent: aBlock
	"Answer the production named aSymbol, if there is no such production answer the result of evaluating aBlock."
	
	(self class ignoredNames includes: aSymbol asString)
		ifTrue: [ ^ aBlock value ].
	(self class startSymbol = aSymbol)
		ifTrue: [ ^ parser ].
	^ self instVarAt: (self class allInstVarNames
		indexOf: aSymbol asString
		ifAbsent: [ ^ aBlock value ])
!   !
!PPCompositeParser methodsFor: #accessing! 
start
	"Answer the production to start this parser with."
	
	self subclassResponsibility
! !
!PPUnresolvedParser methodsFor: #testing!  
isUnresolved
	^ true
! !
!PPUnresolvedParser methodsFor: #parsing!  
parseOn: aStream
	self error: self printString , ' need to be resolved before execution.'
!    !
